-- 4. Huffman-Bume
module Huffman where

import BinTree


-- Datentyp Bit

data Bit = L | R
	 deriving(Eq,Show)

type CodeTable = [(Char,[Bit])]
type SampleList = [(Char,Int)]
data HuffTree = Tip Int Char
	      | Node Int HuffTree HuffTree
		deriving(Show)

-- Kodieren und Dekodieren

decode :: BTree Char -> [Bit] -> [Char]
decode t cs 
    = if null cs then [] else decode1 t cs
    where decode1 (Leaf x) cs = x : decode t cs
	  decode1 (Fork xt yt) (L:cs) = decode1 xt cs
	  decode1 (Fork xt yt) (R:cs) = decode1 yt cs

encode :: BTree Char -> [Char] -> [Bit]
encode t = concat . map (codelookup codetable)
	   where codetable = transform t

codelookup :: CodeTable -> Char -> [Bit]
codelookup ((x,bs) : xbs) y = if x == y then bs else codelookup xbs y

transform :: BTree Char -> CodeTable
transform (Leaf x) = [(x,[])]
transform (Fork xt yt) = hufmerge (transform xt) (transform yt)

hufmerge :: CodeTable -> CodeTable -> CodeTable
hufmerge [] ycs = [(y,R: cs) | (y,cs) <- ycs]
hufmerge xbs [] = [(x,L: bs) | (x,bs) <- xbs]
hufmerge ((x,bs):xbs) ((y,cs):ycs)
	     | length bs <= length cs = (x,L:bs) : (hufmerge xbs ((y,cs):ycs))
	     | otherwise              = (y,R:cs) : hufmerge ((x,bs):xbs) ycs


-- Analyse des Ursprungstextes

sample :: [Char] -> SampleList
sample = sortby freq . collate . sortby id

collate :: [Char] -> SampleList
collate [] = []
collate (x:xs) = (x,1+(length ys)):collate zs
		 where (ys,zs) = span (==x) xs

myspan :: (a->Bool) -> [a] -> ([a],[a])
myspan p [] = ([],[])
myspan p (x:xs) = if p x then (x : ys, zs) else ([],x:xs)
		  where (ys,zs) = span p xs

freq :: (Char,Int) -> Int
freq (x,m) = m

-- sortby aus Kapitel 5. Ineffizient, da Laufzeit ~ n*n

sortby :: (Ord b) => (a -> b) -> [a] -> [a]
sortby f = foldr (insertby f) []

insertby :: (Ord b) => (a -> b) -> a -> [a] -> [a]
insertby f x = insert . span test
	       where insert (xs,ys) = xs ++ [x] ++ ys
		     test y = (f y < f x)


-- Aufbau des Huffman-Baumes

unlabel :: HuffTree -> BTree Char
unlabel (Tip w x) = Leaf x
unlabel (Node w xt yt) = Fork (unlabel xt) (unlabel yt)

mkHuff :: SampleList -> HuffTree
mkHuff = unwrap . until singleton combine . map mktip
	 where mktip (c,w) = Tip w c

combine :: [HuffTree] -> [HuffTree]
combine (xt: yt: xts) = insert (Node w xt yt) xts
			where w = weight xt + weight yt

weight :: HuffTree -> Int
weight (Tip w c) = w
weight (Node w x y) = w

singleton :: [a] -> Bool
singleton xs  = not (null xs) && null (tail xs)

unwrap :: [a] -> a
unwrap [x] = x

insert :: HuffTree -> [HuffTree] -> [HuffTree]
insert xt yts = uts ++ [xt] ++ vts
		where (uts,vts) = span p yts
		      p yt = (weight yt <= weight xt)

-- Beispiel

text :: String
text = "text"

samplelist :: SampleList
samplelist = sample text

hufftree :: HuffTree
hufftree = mkHuff samplelist

enc :: [Bit]
enc = encode (unlabel hufftree) text

dec :: String
dec = decode (unlabel hufftree) enc